#!/usr/bin/env python3
import urllib.request
import datetime
import bs4
import re

url_prefix = 'https://www.berlin.de'
pm_dir = '/sen/gpg/service/presse/2020/'
pm_nav_path = pm_dir + '?page_at_1_0='

# Map abbreviations to full names (and their alternate spellings).
abbrevs = {
  'CW': {'Charlottenburg-Wilmersdorf'},
  'FK': {'Friedrichshain-Kreuzberg'},
  'Li': {'Lichtenberg'},
  'MH': {'Marzahn-Hellersdorf'},
  'Mi': {'Mitte'},
  'Ne': {'Neukölln', 'Neuköln'},
  'Pa': {'Pankow'},
  'Re': {'Reinickendorf'},
  'Sp': {'Spandau'},
  'SZ': {'Steglitz-Zehlendorf'},
  'TS': {'Tempelhof-Schöneberg'},
  'TK': {'Treptow-Köpenick'},
  'sum': {'Summe', 'Berlin'},
}

# some pre-filled values
data = {
   # For these, only image files are available for the table data.
    datetime.datetime(2020, 7, 2): {
        'CW': {'growth': 4, 'total': 851},
        'FK': {'growth': 10, 'total': 681},
        'Li': {'growth': 3, 'total': 427},
        'MH': {'growth': 4, 'total': 468},
        'Mi': {'growth': 0, 'total': 1202},
        'Ne': {'growth': 7, 'total': 1031},
        'Pa': {'growth': 3, 'total': 784},
        'Re': {'growth': 6, 'total': 660},
        'Sp': {'growth': 3, 'total': 450},
        'SZ': {'growth': 0, 'total': 591},
        'TS': {'growth': 3, 'total': 798},
        'TK': {'growth': 0, 'total': 401},
        'sum': {'growth': 43, 'total': 8344}
    },
    datetime.datetime(2020, 4, 5): {
        'CW': {'growth': 9, 'total': 462},
        'FK': {'growth': 2, 'total': 352},
        'Li': {'growth': 0, 'total': 142},
        'MH': {'growth': 3, 'total': 127},
        'Mi': {'growth': 14, 'total': 537},
        'Ne': {'growth': 0, 'total': 392},
        'Pa': {'growth': 10, 'total': 378},
        'Re': {'growth': 9, 'total': 248},
        'Sp': {'growth': 3, 'total': 150},
        'SZ': {'growth': 0, 'total': 312},
        'TS': {'growth': 8, 'total': 394},
        'TK': {'growth': 3, 'total': 193},
        'sum': {'growth': 61, 'total': 3687}
    },
    datetime.datetime(2020, 4, 4): {
        'CW': {'growth': 2, 'total': 453},
        'FK': {'growth': 7, 'total': 350},
        'Li': {'growth': 0, 'total': 142},
        'MH': {'growth': 15, 'total': 124},
        'Mi': {'growth': 22, 'total': 523},
        'Ne': {'growth': 15, 'total': 392},
        'Pa': {'growth': 10, 'total': 368},
        'Re': {'growth': 5, 'total': 239},
        'Sp': {'growth': 21, 'total': 147},
        'SZ': {'growth': 12, 'total': 312},
        'TS': {'growth': 24, 'total': 386},
        'TK': {'growth': 7, 'total': 190},
        'sum': {'growth': 140, 'total': 3626}
    },
    datetime.datetime(2020, 4, 3): {
        'CW': {'growth': 44, 'total': 451},
        'FK': {'growth': 17, 'total': 343},
        'Li': {'growth': 7, 'total': 142},
        'MH': {'growth': 4, 'total': 109},
        'Mi': {'growth': 4, 'total': 501},
        'Ne': {'growth': 40, 'total': 377},
        'Pa': {'growth': 39, 'total': 358},
        'Re': {'growth': 26, 'total': 234},
        'Sp': {'growth': 9, 'total': 126},
        'SZ': {'growth': 18, 'total': 300},
        'TS': {'growth': 41, 'total': 362},
        'TK': {'growth': 14, 'total': 183},
        'sum': {'growth': 263, 'total': 3486}
    },
   # Here the totals needed to be reconstructed.
   datetime.datetime(2020, 3, 9): {
        'CW': {'growth': 4, 'total': 13},
        'FK': {'growth': 3, 'total': 12},
        'Li': {'growth': 0, 'total': 1},
        'MH': {'growth': 1, 'total': 2},
        'Mi': {'growth': 0, 'total': 8},
        'Ne': {'growth': 1, 'total': 3},
        'Pa': {'growth': 1, 'total': 6},
        'Re': {'growth': 0, 'total': 3},
        'Sp': {'growth': 0, 'total': 2},
        'SZ': {'growth': 0, 'total': 3},
        'TS': {'growth': 0, 'total': 5},
        'TK': {'growth': 0, 'total': 0},
        'sum': {'growth': 10, 'total': 58}
   },
   # Here the growth numbers needed to be reconstructed.
   datetime.datetime(2020, 3, 8): {
        'CW': {'growth': 0, 'total': 9},
        'FK': {'growth': 4, 'total': 9},
        'Li': {'growth': 1, 'total': 1},
        'MH': {'growth': 0, 'total': 1},
        'Mi': {'growth': 0, 'total': 8},
        'Ne': {'growth': 0, 'total': 2},
        'Pa': {'growth': 0, 'total': 5},
        'Re': {'growth': 0, 'total': 3},
        'Sp': {'growth': 2, 'total': 2},
        'SZ': {'growth': 1, 'total': 3},
        'TS': {'growth': 0, 'total': 5},
        'TK': {'growth': 0, 'total': 0},
        'sum': {'growth': 8, 'total': 48}
   },
   # Here the growth numbers needed to be reconstructed.
   datetime.datetime(2020, 3, 7): {
        'CW': {'growth': 6, 'total': 9},
        'FK': {'growth': 1, 'total': 5},
        'Li': {'growth': 0, 'total': 0},
        'MH': {'growth': 0, 'total': 1},
        'Mi': {'growth': 1, 'total': 8},
        'Ne': {'growth': 0, 'total': 2},
        'Pa': {'growth': 1, 'total': 5},
        'Re': {'growth': 0, 'total': 3},
        'Sp': {'growth': 0, 'total': 0},
        'SZ': {'growth': 2, 'total': 2},
        'TS': {'growth': 1, 'total': 5},
        'TK': {'growth': 0, 'total': 0},
        'sum': {'growth': 12, 'total': 40}
   },
   # Here the growth numbers needed to be reconstructed.
   datetime.datetime(2020, 3, 6): {
        'CW': {'growth': 1, 'total': 3},
        'FK': {'growth': 0, 'total': 4},
        'Li': {'growth': 0, 'total': 0},
        'MH': {'growth': 0, 'total': 1},
        'Mi': {'growth': 4, 'total': 7},
        'Ne': {'growth': 1, 'total': 2},
        'Pa': {'growth': 1, 'total': 4},
        'Re': {'growth': 0, 'total': 3},
        'Sp': {'growth': 0, 'total': 0},
        'SZ': {'growth': 0, 'total': 0},
        'TS': {'growth': 2, 'total': 4},
        'TK': {'growth': 0, 'total': 0},
        'sum': {'growth': 9, 'total': 28}
   },
   # Here the growth numbers needed to be reconstructed.
   datetime.datetime(2020, 3, 5): {
        'CW': {'growth': 2, 'total': 2},
        'FK': {'growth': 0, 'total': 4},
        'Li': {'growth': 0, 'total': 0},
        'MH': {'growth': 0, 'total': 1},
        'Mi': {'growth': 0, 'total': 3},
        'Ne': {'growth': 0, 'total': 1},
        'Pa': {'growth': 1, 'total': 3},
        'Re': {'growth': 2, 'total': 3},
        'Sp': {'growth': 0, 'total': 0},
        'SZ': {'growth': 0, 'total': 0},
        'TS': {'growth': 1, 'total': 2},
        'TK': {'growth': 0, 'total': 0},
        'sum': {'growth': 6, 'total': 19}
   },
   # Here the growth numbers needed to be reconstructed.
   datetime.datetime(2020, 3, 4): {
        'CW': {'growth': 0, 'total': 0},
        'FK': {'growth': 2, 'total': 4},
        'Li': {'growth': 0, 'total': 0},
        'MH': {'growth': 0, 'total': 1},
        'Mi': {'growth': 0, 'total': 3},
        'Ne': {'growth': 0, 'total': 1},
        'Pa': {'growth': 1, 'total': 2},
        'Re': {'growth': 1, 'total': 1},
        'Sp': {'growth': 0, 'total': 0},
        'SZ': {'growth': 0, 'total': 0},
        'TS': {'growth': 0, 'total': 1},
        'TK': {'growth': 0, 'total': 0},
        'sum': {'growth': 4, 'total': 13}
   },
   # Here the growth numbers needed to be reconstructed.
   datetime.datetime(2020, 3, 3): {
        'CW': {'growth': 0, 'total': 0},
        'FK': {'growth': 2, 'total': 2},
        'Li': {'growth': 0, 'total': 0},
        'MH': {'growth': 0, 'total': 1},
        'Mi': {'growth': 0, 'total': 3},
        'Ne': {'growth': 0, 'total': 1},
        'Pa': {'growth': 1, 'total': 1},
        'Re': {'growth': 0, 'total': 0},
        'Sp': {'growth': 0, 'total': 0},
        'SZ': {'growth': 0, 'total': 0},
        'TS': {'growth': 0, 'total': 1},
        'TK': {'growth': 0, 'total': 0},
        'sum': {'growth': 3, 'total': 9}
   },
}
fixes = {
   # Here the official total is 215, while the summation of district
   # numbers only adds up to 125 – pretty much looks like a mere
   # transposition of digits.
   datetime.datetime(2020, 3, 27): {
       'sum': {
           'growth': 125
       }
   },
   # Here the official total is 1937, while the summation of district
   # numbers only adds up to 1917; furthermore, the original value for
   # SZ is 118 (+18), which makes no sense, as the day before is
   # 120 (+15) and the day after is 147 (+15).  The following is a
   # compromise to keep as many surrounding numbers stable as possible.
   datetime.datetime(2020, 3, 26): {
       'SZ': {
           'growth': 12
       },
       'sum': {
           'growth': 286
       }
   },
   # Here the official total is 220, while the summation of district
   # numbers adds up to 228 – looks like someone misread an 8 as a 0.
   datetime.datetime(2020, 3, 25): {
       'sum': {
           'growth': 220
       }
   },
}

# Scan navigation bar for maximum pagination value.
url = url_prefix + pm_dir
with urllib.request.urlopen(url) as response:
   html = response.read()
soup = bs4.BeautifulSoup(html, 'html.parser')
max_page=0
for link in soup.find_all('a'):
    href = link['href']
    if str.startswith(href, pm_nav_path):
        max_test = int(href.split('=')[1])
        max_page = max_test if max_test > max_page else max_page

# Scan paginated press release links for daily Corona number briefing links.
day_urls = []
for i in range(max_page):
    url = url_prefix + pm_nav_path + str(i + 1)
    with urllib.request.urlopen(url) as response:
        html = response.read()
    soup = bs4.BeautifulSoup(html, 'html.parser')
    for link in soup.find_all('a'):
        if (not link.string) or\
           (not link.string.startswith('Coronavirus: Derzeit') and
            not link.string.startswith('Coronavirus in Berlin: Bestätigte Fälle')):
            continue
        day_urls += [link['href']]

# Collect infection data.
first_run = True
districts_sorted = []
date_limit = datetime.datetime(2020, 3, 11)
for path in day_urls:
    url = url_prefix + path
    with urllib.request.urlopen(url) as response:
        html = response.read()
    soup = bs4.BeautifulSoup(html, 'html.parser')
    date_title = soup.find('div', class_='pressnumber')
    m = re.search('[0-9]+\\.[0-9]+\\.[0-9]+', date_title.string)
    date_formatted = m.group(0)
    date = datetime.datetime.strptime(date_formatted , '%d.%m.%Y')
    if date_limit > date:
        break
    # On that day, two press releases were released, for that and the prev day.
    if date == datetime.datetime(2020, 3, 15) and date in data:
       date = datetime.datetime(2020, 3, 14)
    # From here on, press releases describe numbers from prev day.
    if date <= datetime.datetime(2020, 3, 13):
       date = date - datetime.timedelta(days=1)
    table = soup.find('table')
    # For 13th of March we lack a press release.
    if table is None and (date in data or date == datetime.datetime(2020, 3, 13)):
        continue
    data[date] = {}
    for tr in [tr for tr in table.children if type(tr) == bs4.element.Tag][1:]:
         printable_tds = []
         for td in [td for td in tr.children if type(td) == bs4.element.Tag][:2]:
             printable_string = ' '.join([s for s in td.strings])
             printable_tds += [printable_string.strip()]
         district_long = printable_tds[0]
         district_short = [k for k in abbrevs if district_long in abbrevs[k]][0]
         if first_run:
             districts_sorted += [district_short]
         if date == datetime.datetime(2020, 3, 10):
            # For this date we only get totals.
            data[date][district_short] = {'total': int(printable_tds[1])}
         else:
             split_char = ' '
             if not split_char in printable_tds[1]:
                 split_char = '('
             total_str, growth_str = printable_tds[1].split(split_char)
             growth = int(growth_str.replace('(', '').replace(')', '').\
                          replace('+', ''))
             total = int(total_str.replace('.', ''))
             data[date][district_short] = {'growth': growth, 'total': total}
    first_run = False

def neighbor_days(day_target):
   day_delta = datetime.timedelta(days=1)
   return day_target + day_delta, day_target - day_delta

# Reconstruct growth for 10th of March.
day_target = datetime.datetime(2020, 3, 10)
day_after, day_before = neighbor_days(day_target)
for district in [d for d in districts_sorted]:
   total_target = data[day_target][district]['total']
   total_before = data[day_before][district]['total']
   data[day_target][district]['growth'] = total_target - total_before

# Reconstruct data for 13th of March.
day_target = datetime.datetime(2020, 3, 13)
day_after, day_before = neighbor_days(day_target)
data[day_target] = {}
for district in [d for d in districts_sorted]:
   data[day_target][district] = {}
   total_after = data[day_after][district]['total']
   growth_after = data[day_after][district]['growth']
   total_target = total_after - growth_after
   data[day_target][district]['total'] = total_target
   total_before = data[day_before][district]['total']
   data[day_target][district]['growth'] = total_target - total_before

dates_sorted = list(data.keys())
dates_sorted.sort()
dates_sorted.reverse()

# Apply fixes and ensure integrity of results
for date in fixes:
    for district in fixes[date]:
        for type_ in fixes[date][district]:
            data[date][district][type_] = fixes[date][district][type_]
for date in dates_sorted:
    if date in fixes:
       continue
    for district in [d for d in districts_sorted if not d=='sum']:
        prev_date = date - datetime.timedelta(days=1)
        if prev_date not in dates_sorted:
           if prev_date >= date_limit:
              raise Exception('Dates not contiguous: %s missing', prev_date)
           else:
              continue
        prev_total = data[prev_date][district]['total']
        cur_total = data[date][district]['total']
        if cur_total - data[date][district]['growth'] != prev_total:
            raise Exception('Questionable district infection total in %s/%s' % (district, date))
    day_sum = 0
    for district in [d for d in districts_sorted if not d=='sum']:
       day_sum += data[date][district]['total']
    if day_sum != data[date]['sum']['total']:
        raise Exception('Questionable district infection total sum in %s' % date)
    day_sum = 0
    for district in [d for d in districts_sorted if not d=='sum']:
       day_sum += data[date][district]['growth']
    if day_sum != data[date]['sum']['growth']:
        raise Exception('Questionable district infection growth sum in %s' % date)

# Final output.
dates_sorted.reverse()
print(' '*10, ' '.join(['%3s' % d for d in districts_sorted]))
for date in dates_sorted:
    growths = []
    for d in districts_sorted:
        growths += [data[date][d]['growth']]
    print(date.strftime('%Y-%m-%d'), ' '.join(['%3s' % g for g in growths]))
